package partnerpay

import (
	"bytes"
	"context"
	"crypto/rsa"
	"errors"
	"fmt"
	"github.com/wechatpay-apiv3/wechatpay-go/core"
	"github.com/wechatpay-apiv3/wechatpay-go/core/auth/verifiers"
	"github.com/wechatpay-apiv3/wechatpay-go/core/downloader"
	"github.com/wechatpay-apiv3/wechatpay-go/core/notify"
	"github.com/wechatpay-apiv3/wechatpay-go/core/option"
	"github.com/wechatpay-apiv3/wechatpay-go/services/partnerpayments"
	"github.com/wechatpay-apiv3/wechatpay-go/services/partnerpayments/jsapi"
	"github.com/wechatpay-apiv3/wechatpay-go/services/payments"
	"github.com/wechatpay-apiv3/wechatpay-go/utils"
	"log"
	"math/big"
	"net/http"
	"strings"
	"time"
)

// WechatPartnerPayClient 服务商支付
type WechatPartnerPayClient struct {
	AppId         *string         `json:"appId,omitempty"`         // 微信小程序Id
	SpMchId       *string         `json:"spMchId,omitempty"`       // 服务商商户号(如果服务商模式，该值需要填写)
	MchId         *string         `json:"mchId,omitempty"`         // 收款商户号(如果是服务商模式，配置特约商户号)
	MchSerialNo   *string         `json:"mchSerialNo,omitempty"`   // 收款商户证书序号
	MchPrivateKey *string         `json:"mchPrivateKey,omitempty"` // 收款商户私钥
	ApiKey        *string         `json:"apiKey,omitempty"`        // apiKey
	wxClient      *core.Client    //不需要外界传入
	privateKey    *rsa.PrivateKey //不需要外界传入
}

// WxJsApiOrder JSAPI下单
func (s *WechatPartnerPayClient) WxJsApiOrder(p *PaymentJsApiParam) *PaymentJsApiResponse {
	wxClient := s.getWxPayClient()
	svc := jsapi.JsapiApiService{Client: wxClient}

	money, b := new(big.Int).SetString(new(big.Float).Mul(new(big.Float).SetFloat64(*p.Money), new(big.Float).SetFloat64(100)).String(), 10)
	if !b {
		panic(errors.New("金额错误请检查，请检查"))
	}
	pr := jsapi.PrepayRequest{
		SpAppid:     s.AppId,
		SpMchid:     s.SpMchId,
		SubMchid:    s.MchId,
		Description: p.Title,
		OutTradeNo:  p.OutTradeNo,
		NotifyUrl:   p.CallBackUrl,
		TimeExpire:  core.Time(getTimeExpire(p.TimeExpire)),
		Amount: &jsapi.Amount{
			Total: core.Int64(money.Int64()),
		},
		Payer: &jsapi.Payer{
			SpOpenid: p.OpenId,
		},
		Attach: p.Attach,
	}
	resp, _, err := svc.Prepay(context.Background(), pr)
	if err != nil {
		panic(errors.New(getErrMessage(err.Error())))
	}
	log.Println(fmt.Sprintf("微信调用结果：%s", resp.String()))

	return buildWxMiniPayInfo(*resp.PrepayId, *s.AppId, s.privateKey)
}

func (s *WechatPartnerPayClient) getWxPayClient() *core.Client {
	if s.wxClient != nil {
		return s.wxClient
	}
	// 使用 utils 提供的函数从本地文件中加载商户私钥，商户私钥会用来生成请求的签名
	if s.MchPrivateKey == nil {
		panic(errors.New("商户私钥未配置，请检查"))
	}
	mchPrivateKey, err := LoadPrivateKey(*s.MchPrivateKey)
	if err != nil {
		panic(errors.New("商户私钥配置错误，请检查"))
	}
	s.privateKey = mchPrivateKey

	// 使用商户私钥等初始化 client，并使它具有自动定时获取微信支付平台证书的能力
	opts := []core.ClientOption{
		option.WithWechatPayAutoAuthCipher(*s.SpMchId, *s.MchSerialNo, mchPrivateKey, *s.ApiKey),
	}
	wc, err := core.NewClient(context.Background(), opts...)
	if err != nil {
		panic(err)
	}
	s.wxClient = wc
	return wc
}

// AnalysisPayNotifyResult 解析支付通知结果
// https://pay.weixin.qq.com/wiki/doc/apiv3_partner/apis/chapter4_5_5.shtml
func (s *WechatPartnerPayClient) AnalysisPayNotifyResult(request *http.Request) (*notify.Request, *payments.Transaction, error) {
	certVisitor := downloader.MgrInstance().GetCertificateVisitor(*s.SpMchId)
	handler, e := notify.NewRSANotifyHandler(*s.ApiKey, verifiers.NewSHA256WithRSAVerifier(certVisitor))
	if e != nil {
		return nil, nil, e
	}
	content := new(payments.Transaction)
	notifyReq, err := handler.ParseNotifyRequest(context.Background(), request, content)
	if err != nil {
		return nil, nil, errors.New(getErrMessage(err.Error()))
	}
	return notifyReq, content, nil
}

// QueryOrderByOutTradeNo 商户订单号查询订单
// https://pay.weixin.qq.com/wiki/doc/apiv3_partner/apis/chapter4_5_2.shtml
func (s *WechatPartnerPayClient) QueryOrderByOutTradeNo(payCode string) (*partnerpayments.Transaction, error) {
	wxClient := s.getWxPayClient()
	svc := jsapi.JsapiApiService{Client: wxClient}
	resp, _, err := svc.QueryOrderByOutTradeNo(context.Background(),
		jsapi.QueryOrderByOutTradeNoRequest{
			OutTradeNo: &payCode,
			SpMchid:    s.SpMchId,
			SubMchid:   s.MchId})
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func getTimeExpire(expireTime *int64) time.Time {
	if expireTime == nil || *expireTime == 0 {
		t := defaultExpireTime
		expireTime = &t
	}
	return dateAdd(time.Now(), *expireTime*60)
}

func dateAdd(date time.Time, d int64) time.Time {
	t := date.Unix()
	t += d //增加秒
	return time.Unix(t, 0)
}

func getErrMessage(msg string) string {
	startStr := "response:["
	endStr := "Header"
	startIndex := strings.Index(msg, startStr)
	endIndex := strings.Index(msg, endStr)
	rs := msg
	if startIndex >= 0 && endIndex > 0 {
		rs = msg[startIndex+len(startStr) : endIndex]
	}
	return strings.ReplaceAll(rs, "\n", " ")
}

func buildWxMiniPayInfo(prepayId, appId string, privateKey *rsa.PrivateKey) *PaymentJsApiResponse {
	nonceStr, err := utils.GenerateNonce()
	if err != nil {
		panic(err)
	}
	packageInfo := "prepay_id=" + prepayId
	r := &PaymentJsApiResponse{
		TimeStamp: core.Int64(time.Now().Unix()),
		NonceStr:  &nonceStr,
		Package:   core.String("prepay_id=" + prepayId),
		SignType:  core.String("RSA"),
	}
	r.PaySign = rsaSign(appId, *r.TimeStamp, nonceStr, packageInfo, privateKey)
	return r
}

func rsaSign(appId string, timeStamp int64, nonceStr, packageInfo string, privateKey *rsa.PrivateKey) *string {
	var buf bytes.Buffer
	buf.WriteString(fmt.Sprintf("%s\n", appId))
	buf.WriteString(fmt.Sprintf("%d\n", timeStamp))
	buf.WriteString(fmt.Sprintf("%s\n", nonceStr))
	buf.WriteString(fmt.Sprintf("%s\n", packageInfo))
	sign, err := utils.SignSHA256WithRSA(buf.String(), privateKey)
	if err != nil {
		panic(err)
	}
	return &sign
}
